@@ -1,6 +1,6 @@
 /* A more useful interface to strtol.
 
-   Copyright (C) 1995, 1996, 1998, 1999, 2001, 2002, 2003, 2004, 2006
+   Copyright (C) 1995, 1996, 1998, 1999, 2001, 2002, 2003, 2004, 2006, 2007
    Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
@@ -20,8 +20,7 @@
 #ifndef XSTRTOL_H_
 # define XSTRTOL_H_ 1
 
-# include "exitfail.h"
-
+# include <getopt.h>
 # include <inttypes.h>
 
 # ifndef _STRTOL_ERROR
@@ -48,37 +47,33 @@ _DECLARE_XSTRTOL (xstrtoul, unsigned long int)
 _DECLARE_XSTRTOL (xstrtoimax, intmax_t)
 _DECLARE_XSTRTOL (xstrtoumax, uintmax_t)
 
-# define _STRTOL_ERROR(Exit_code, Str, Argument_type_string, Err)	\
-  do									\
-    {									\
-      switch ((Err))							\
-	{								\
-	default:							\
-	  abort ();							\
-									\
-	case LONGINT_INVALID:						\
-	  error ((Exit_code), 0, "invalid %s `%s'",			\
-		 (Argument_type_string), (Str));			\
-	  break;							\
-									\
-	case LONGINT_INVALID_SUFFIX_CHAR:				\
-	case LONGINT_INVALID_SUFFIX_CHAR | LONGINT_OVERFLOW:		\
-	  error ((Exit_code), 0, "invalid character following %s in `%s'", \
-		 (Argument_type_string), (Str));			\
-	  break;							\
-									\
-	case LONGINT_OVERFLOW:						\
-	  error ((Exit_code), 0, "%s `%s' too large",			\
-		 (Argument_type_string), (Str));			\
-	  break;							\
-	}								\
-    }									\
-  while (0)
+#ifndef __attribute__
+# if __GNUC__ < 2 || (__GNUC__ == 2 && __GNUC_MINOR__ < 8) || __STRICT_ANSI__
+#  define __attribute__(x)
+# endif
+#endif
 
-# define STRTOL_FATAL_ERROR(Str, Argument_type_string, Err)		\
-  _STRTOL_ERROR (exit_failure, Str, Argument_type_string, Err)
+#ifndef ATTRIBUTE_NORETURN
+# define ATTRIBUTE_NORETURN __attribute__ ((__noreturn__))
+#endif
 
-# define STRTOL_FAIL_WARN(Str, Argument_type_string, Err)		\
-  _STRTOL_ERROR (0, Str, Argument_type_string, Err)
+/* Report an error for an invalid integer in an option argument.
+
+   ERR is the error code returned by one of the xstrto* functions.
+
+   Use OPT_IDX to decide whether to print the short option string "C"
+   or "-C" or a long option string derived from LONG_OPTION.  OPT_IDX
+   is -2 if the short option "C" was used, without any leading "-"; it
+   is -1 if the short option "-C" was used; otherwise it is an index
+   into LONG_OPTIONS, which should have a name preceded by two '-'
+   characters.
+
+   ARG is the option-argument containing the integer.
+
+   After reporting an error, exit with a failure status.  */
+
+void xstrtol_fatal (enum strtol_error,
+		    int, char, struct option const *,
+		    char const *) ATTRIBUTE_NORETURN;
 
 #endif /* not XSTRTOL_H_ */
